#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from .model_deployment import ModelDeployment
from .model_deployer import ModelDeployer
from .model_deployment_properties import ModelDeploymentProperties
